/*
 * libkarma/riocp.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004-2005
 *               Enrique Vidal 2004-2005
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef LINUX
#include <ftw.h>
#else
#include "ftw.h"
#endif
#include <locale.h>
#include <langinfo.h>

#include "karma.h"
#include "ssdp.h"
#include "properties.h"
#include "util.h"
#include "errors.h"
#include "rio_rw.h"

#define CHECK(x) (x)?"failed":"ok"

#define defaultCodeset "ISO8859-15"

#define ssdpRetry 3                     /* number of retries allowed for SSDP */

void usage(void);
void currentKarma(int i, int *k);
int treecopy(const char *file, const struct stat *sb, int flag);
uint32_t * searchStr(char *searchString);
double getUsedSize(int karma);

void fidNamesFileHandler(const int op, const int fid, const char * data);
void deleteFids(uint32_t *vp);
int isodigit (char d);
char esc(char c);
char *expand(char *str);
void tr(char *str, char *orig, char *sub, char *del, int squeezeRep);
void downloadFids(uint32_t *vp, int doDownload,
		char* useTuneProps, char* useTaxiProps, char* usePlstProps,
		char* edPathStr1, char* edPathStr2, char* edPathStr3,
		int squeezeRep);


void usage(void)
{
  printf("usage: riocp [-a {hostname|ip}] [-p password] [-c charset]\n"
         "             [-f [fidNameFile]] [-s searchString ]\n"
	 "             [-u props] [-x props] [-l props] [-e set1 set2 set3\n" 
         "             [-q] [-d] [-r] [-R] [files_to_upload ...]\n"
	 "\n");
  printf("  \"-f\" download fids with filenames given in fidNamesFile,\n"
	 "       where fidNamesFile defaults to ~/.openrio/.../fidNamesFile\n"
         "\n");
  printf("  \"-s\" AND/OR-search karma DB according to a \"searchString\";\n"
	 "       \"searchString\" syntax: <property>=<searchStr>[{,|+}...], "
                                                              "where: \n"
	 "       - <searchStr> is a substring of the <property> value\n"
	 "         (see ~/.openrio/*/fileinfo for possible <property> names)\n"
	 "       - \',\' and \'+\' mean AND and OR, respectively\n"
	 "       - each AND/OR operator is applied to the "
	                        "previous search results\n"
         "\n");
  printf("  \"-u\" use properties to build download path for <type>=\'tune\'. "
	                                            "Default=\'%%f.%%c\'\n"
         "  \"-x\" like -u for <type>=\'taxi\',.    Default=\"-u props\"\n"
         "  \"-l\" like -u for <type>=\'playlist\'. Default=\"-x props\"\n");
  printf("       \"props\" are formating strings (as in \"printf\"), where\n"
         "       \'%%T\', \'%%g\', \'%%a\', \'%%s\', \'%%f\', \'%%t\', "
                                  "\'%%n\', \'%%c\' are, resp., values of\n"
	 "       <type>, <genre>, <artist>, <source>, <fid>, <title>, "
                                                 "<tracknr>, <codec>\n"
	 "\n");
  printf("  \"-e\" edit the download pathname using \"set1\" and \"set2\" "
                                                     "as in unix \'tr\'\n"
         "       - if \"set3\" is empty (\"\"): just translate as in \'tr\';\n"
	 "       - otherwise, further delete all chars given in \"set3\"\n"
	 "  \"-q\" squeeze repeated chars (given in \"set2\"),"
                                         " before the optional delete\n\n");
  printf("  \"-d\" download the files searched for with \"-s\"\n"
         "  \"-r\" remove these files (after the optional download)\n"
         "  \"-R\" same as -r, but do not ask for confirmation\n");
}

void currentKarma(int i, int *k) 
{
	static int karma;
	if (i)
		karma=*k;
	else
		*k=karma;
}

void fidNamesFileHandler(const int op, const int fid, const char * data) {
  char *serial, *fidNamesFileName;
  static FILE *fidNames;
  int karma, fid1;
  char filename[256]="";

/*fprintf(stderr, "-->%d, %d, %s\n", op, fid, data);*/
  if (op==0) {                                                       /* Close */
    if(fidNames) fclose(fidNames);
  }
  else if ((op==1) || (op==2)) {                        /* Read or Write Open */
    if ((data!=NULL) && (strlen(data)>0))
      if (fidNames) fidNamesFileName=strdup(data);

    fidNamesFileName=strdup(getenv("HOME"));
    /*33=strlen("/.openrio/pearl-0123456/fileinfo")*/
    fidNamesFileName=realloc(fidNamesFileName, strlen(fidNamesFileName)+37);
    serial = (char *)calloc(8, sizeof(char));
    snprintf(serial, 7, "%d", atoi(get_property(0,"serial")));
    strcat(fidNamesFileName,"/.openrio/");
    strcat(fidNamesFileName,"pearl-");
    strcat(fidNamesFileName,serial);
    free(serial);
    strcat(fidNamesFileName, "/");
    strcat(fidNamesFileName,"fidNamesFile");
    /*  fprintf(stderr, "fidNamesFileName = %s\n", fidNamesFileName); */
    if(create_path(fidNamesFileName)==-1){
	    return;
    }
    fidNames = fopen(fidNamesFileName, (op==1)?"r":"a");    /* append writing */
    if (fidNames==NULL)
	    fprintf(stderr, "** Error: can't open %s\n", fidNamesFileName);
    free(fidNamesFileName);
  }
  else if (op==3) {                    /* Download files as given in fidNames */
	  currentKarma(0, &karma);
	  while(fscanf(fidNames, "%d %s\n", &fid1, filename) > 0) {
		  printf("Downloading fid %i to %s\n", fid1, filename);
		  fprintf(stderr, "rio_read %s\n", CHECK(rio_read(karma, fid1, filename)));
		  filename[0]='\0';
	  }
  }
  else if (op==4) {            /* Write fid and filename=data to fidNamesFile */
	  if (fprintf(fidNames, "%d %s\n", fid, data)<0)
		  fprintf(stderr, "** Error writing to fidNamesFile\n");
  }
  else if (op==5) {}                        /* Upload   fidNamesFile to   Rio */
  else if (op==6) {}                        /* Download fidNamesFile from Rio */
  else fprintf(stderr, "** Error: unknown op in fidNamesFileHandler\n");
}

int treecopy(const char *file, const struct stat *sb, int flag)
{
	static int first = 1;
	static int karma;
	uint32_t fid;

	if(first){   
		currentKarma(0, &karma);
		first=0;
	}
	if(flag==FTW_F) {
		fid = rio_write(karma, file);
		printf("Uploading fid %d, %s: %s\n", 
                                  fid, file, (fid>0)?"ok":"failed");
 		fidNamesFileHandler(4, fid, file);
	}
	return 0;
}

uint32_t * searchStr(char *searchString)
{
  int c, sizec=0, sizep=-1, andOr=-1;              /* andOr = 1 (and), 0 (or) */
  uint32_t *vc=NULL, *vp=NULL;                    /* vc/vp : current/previous */
  char *keyValue=NULL, *key=NULL;
  char *opp=NULL, op='\0';          /* op: logical operator;  opp: op pointer */

  if (searchString==NULL) return NULL;

  op = ((opp = strpbrk(searchString, ",+"))==NULL) ? '\0' : opp[0];
  keyValue=strsep(&searchString, "+,");    /* searchString -> past "," or "+" */

  while (keyValue!=NULL) {
    fprintf(stderr, "%9s FIDs with \"%-30s:", 
          (andOr==-1)?"Searching":(andOr?"ANDing":"ORing"), keyValue);

    key=strsep(&keyValue, "=");                   /* keyValue -> past the "=" */

    if (sizep<0) vc=andOrSearch(andOr, NULL, key, keyValue);
    else         vc=andOrSearch(andOr, vp,   key, keyValue);    /* frees vp!! */

    c=0; while(vc[c] != 0) c++; sizec=c;                        /* size of vc */
    fprintf(stderr, " %4d selected\n", sizec);

    vp = malloc((sizec+1)*sizeof(uint32_t));
    memcpy(vp, vc, (sizec+1)*sizeof(uint32_t));                 /* copy vp=vc */
    sizep=sizec;
    free(vc);

    if (searchString==NULL) break;

    andOr = (op==',');
    op = ((opp = strpbrk(searchString, ",+"))==NULL) ? '\0' : opp[0];
    keyValue=strsep(&searchString, "+,");  /* searchString -> past "," or "+" */
  }
  return vp;
}

void deleteFids(uint32_t *vp) {
  int karma, c;

  currentKarma(0, &karma);
  for (c=0; vp[c]!=0; c++) 
    if (vp[c]!=256) {                               /* ignoring root playlist */
    	fprintf(stderr, "delete_file(fid=%d): %s\n", 
	             vp[c], CHECK(delete_file(karma, vp[c])));
	delete_property(vp[c]);
    }
}

/* -------------------------------------------------------------------------- */
/* ---------------------------- Pathname editing ---------------------------- */
/* -- Adapted from "tr.c",  Copyright 1998 ams@wiw.org (Abhijit Menon-Sen) -- */

#define ins(p, c) *p++ = c
#define next(p) *p++

int isodigit (char d)
{
    if ('0' <= d && d <= '7') { return 1; }
    return 0;
}

char esc(char c)
{
    switch (c) {
    case 'a': c = '\a'; break;
    case 'b': c = '\b'; break;
    case 'f': c = '\f'; break;
    case 'n': c = '\n'; break;
    case 'r': c = '\r'; break;
    case 't': c = '\t'; break;
    case 'v': c = '\v'; break;
    default:  break;
    }
    return (c);
}

char *expand(char *str)
{
    static char buf[512], num[8];
    char i, j=0, k, *o = buf, *p = num;
    int a, class = 0;

    while ((i = next(str))) {
        switch (i) {
        case '\\': if (!(k = next(str))) {
                       fprintf(stderr, "hanging '\\'.\n");
                       exit(1);
                   }
                   else if (tolower(k) == 'x') {
                       p = num;
                       if ((k = next(str)) == 0 || !isxdigit(k)) {
                           fprintf(stderr, "parse error in \\x.\n");
                           exit(1);
                       }
                       *p++ = k; *p = 0; a = strtol(num, NULL, 16);
                       if ((k = next(str)) == 0 || !isxdigit(k)) {
                           ins(o, (char)a); str--;
                       }
                       else {
                           *p++ = k; *p = 0; a = strtol(num, NULL, 16);
                           ins(o, (char)a);
                       }
                   }
                   else if (isodigit(k)) {
                       p = num;
                       *p++ = k; *p = 0; a = strtol(num, NULL, 8);
                       if ((k = next(str)) == 0 || !isodigit(k)) {
                           ins(o, (char)a); str--;
                       }
                       else {
                           *p++ = k; *p = 0; a = strtol(num, NULL, 8);
                           if ((k = next(str)) == 0 || !isodigit(k)) {
                               ins(o, (char)a); str--;
                           }
                           else {
                               *p++ = k; *p = 0; a = strtol(num, NULL, 8);
                               ins(o, (char)a);
                           }
                       }
                   }
                   else { ins(o, esc(k)); }
                   break;
        case '[':  if (++class > 1) {
                       fprintf(stderr, "* Literal [ inside character class.\n");
                       exit(1);
                   }
                   break;
        case '.':  if (!class) { ins(o, '.'); }
                   else {
                       for (a = 0x20; a < 0x7f; ++a) { ins(o, (char)a); }
                   }
                   break;
        case '-':  if (!class || (k = next(str)) == ']') {
                       ins(o, '-');
                       if (class) { class--; }
                   }
                   else {
                       if (j < k) {
                           for (i = j+1; i <= k; ++i) { ins(o, i); }
                       }
                       else if (k < j) {
                           for (i = j-1; i >= k; --i) { ins(o, i); }
                       }
                   }
                   break;
        case ']':  if (--class) {
                       fprintf(stderr,"* literal ] outside character class.\n");
                       exit(1);
                   }
                   break;
        default:   ins(o, i);
        }
        j = i;
    }
    ins(o, 0);
    return (buf);
}

void tr(char *str, char *orig, char *sub, char *del, int squeezeRep)
{
    char c, a[1024], b[1024], d[1024];
    int tab[256], i, cc, s1, s2;
    char *q, *p=str;

    for (i = 0; i < 256; ++i) {
        tab[i] = i;
    }
    strcpy(a, expand(orig));
    strcpy(b, expand(sub));
    strcpy(d, expand(del));
    s1 = strlen(a);
    s2 = strlen(b);

    for (i = 0; i < s1; ++i) {
        cc=(int)a[i]; cc = (cc<0)?cc+256:cc;
        tab[cc] = i < s2 ? b[i] : b[s2 - 1];
/*      tab[(int)a[i]] = i < s2 ? b[i] : b[s2 - 1]; */
    }
    while ((c = *str)) {
        cc = (c<0)?c+256:c;
        *str++ = tab[cc];
    }
    if (squeezeRep) {
      q=str=p;
      while ((c = *str)) {
	*str++;
	if ((c != *str) || (strchr(b, c)==NULL)) {*q++; *q=*str;}
      }
    }
    if(del && (strlen(del)>0)) {
      q=str=p;
      while ((c = *str)) {
	if ((strchr(d, c)==NULL)) *q++=*str;
	*str++;
      }
      *q='\0';
    }
}

/* -------------------------- End pathname editing -------------------------- */
/* -------------------------------------------------------------------------- */

void downloadFids(uint32_t *vp, int doDownload,
		  char* useTuneProps, char* useTaxiProps, char* usePlstProps, 
                  char* edPathStr1, char* edPathStr2, char* edPathStr3, 
		  int squeezeRep)
{
  int karma, i, j;
  char c, cc, *useProps=NULL, filename[512]="\0";
  char *type, *genre, *artist,*source, *title, *tracknr, *codec;
  uint32_t fid;

  currentKarma(0, &karma);
  for (j=0; vp[j]!=0; j++) {
    if((!doDownload) && (j==0)) 
      printf("# Fid Name\n#---- -------------------------------------\n");
        fid=vp[j];
        type=get_property(fid, "type");
        if(type==NULL) {
          fprintf(stderr, "**Could not get type of fid=%d\n", fid);
          return;
        }
	if(strcmp((char *)type, "playlist") == 0) useProps=strdup(usePlstProps);
	if(strcmp((char *)type, "taxi") == 0)     useProps=strdup(useTaxiProps);
	if(strcmp((char *)type, "tune") == 0)     useProps=strdup(useTuneProps);

	for (i=0; useProps[i]!='\0'; i++) {
	  cc=(useProps[i]);
	  if (cc == '%') {
	    c=useProps[++i];
	    switch(c) {
	    case '\0':
	      fprintf(stderr, "** Format error: %s\n", useProps);
	      i--;
	    break;
	    case '%':
	      strncat(filename, useProps+i, 1);
	    break;
	    case 'T':                                               /* <type> */
	      strcat(filename, type);
	    break;
	    case 'g':                                              /*< genre> */
	      genre=get_property(fid, "genre");
	      if(genre) strcat(filename, genre);
	    break;
	    case 'a':                                             /* <artist> */
	      artist=get_propertyUtf8(fid, "artist");
	      if(artist){
		      strcat(filename, artist);
		      free(artist);
	      }
	    break;
	    case 's':                                     /* <source> (album) */
	      source=get_propertyUtf8(fid, "source");
	      if(source){
		      strcat(filename, source);
		      free(source);
	      }
	    break;
	    case 'f':                                                /* <fid> */
	      strcat(filename, simple_itoa(fid));
	    break;
	    case 't':                                              /* <title> */
	      title=get_propertyUtf8(fid, "title");
	      if(title){
		      strcat(filename, title);
		      free(title);
	      }
	      else{
		fprintf(stderr, "**Could not get title of fid=%d\n", fid);
		/* return; */
	      }
	    break;
	    case 'n':                                            /* <tracknr> */
	      tracknr=get_property(fid, "tracknr");
	      strcat(filename, tracknr);
	    break;
	    case 'c':                                   /* <codec>|"playlist" */
	      if(strcmp((char *)type, "playlist") == 0) 
		strcat(filename, type);
	      else{
		codec=get_property(fid, "codec");
		if(codec) strcat(filename, codec);
		else{
		  fprintf(stderr, "**Could not get codec of fid=%d\n", fid);
		  return;
		}
	      }
	    break;
	    default:
	      fprintf(stderr, "* Unknown property \'%c\' - ignored\n", c);
	    break;
	    } /* switch(c) */
	  }else{                           /* just copy all non-special chars */
	    strncat(filename, useProps+i, 1);
	  } /* if(cc == ...) */
	} /* for i=0 ...  */
	
	free(useProps);

	if(edPathStr1 && edPathStr2 && edPathStr3)   /* 'tr'-editing filename */
	  tr(filename, edPathStr1, edPathStr2, edPathStr3, squeezeRep);

        if(doDownload) {
	  fprintf(stderr, "Downloading fid %i to %s\n", fid, filename);

	  if(strcmp((char *)type, "playlist") == 0) {
	    fprintf(stderr, "rio_read_playlist %s\n",
		             CHECK(rio_read_playlist(karma, fid, filename)));
	  }else{
	    fprintf(stderr, "rio_read %s\n", 
                             CHECK(rio_read(karma, fid, filename)));
	  }
	}else{
	  printf("%5i %s\n", fid, filename);
	}
	filename[0]='\0';
	if((!doDownload) && (vp[j+1]==0)) 
	  printf("#---- -------------------------------------\n");
  }
}

double getUsedSize(int karma)
{
	/* int rio; uint32_t storage_id;  */
	uint32_t n_files, highest_file_id;
        uint64_t s_size, f_space;
	double size, freeSpace;

        n_files=s_size=f_space=highest_file_id=0;
	get_storage_details(karma, 0, 
			    &n_files, &s_size, &f_space, &highest_file_id);
	size = s_size; 
	freeSpace = f_space;
/*      fprintf(stderr, "%E", size - freeSpace); */
	return(size - freeSpace);
}

int main(int argc, char * argv[])
{
	int i, c, r;
        char *settings;
	char *host=NULL;
	char *usercodeset=NULL;/*="ISO8859-15";*/
	char *psw=NULL;    /* "psw" is a bit better to search than old "p" ;) */
	char *searchString=NULL;
	char *tunePropsToUse="%f.%c"; /* props string for tune download paths */
	char *taxiPropsToUse=NULL;    /* props string for taxi download paths */
	char *plstPropsToUse=NULL;    /* props string for playlst downld paths*/
	char *editPathStr1=NULL;
	char *editPathStr2=NULL;
	char *editPathStr3=NULL;
	int  squeezeRepeats=0;
	int  download=0;
	int  fidNamesDownload=0;
	int  removeFiles=0;
	char yesNot='N';
	int  karma;
	uint16_t port;
        uint32_t *vp=NULL;                                  /* vector of fids */
	int dbChanged;

	while ((c = getopt(argc, argv, "dhfqrRa:c:e:p:s:u:x:l:")) != -1) {
		switch(c) {
		case 'h': case '?':
			usage();
			return -1;
			break;
		case 'd':
		        download=1;
			break;
		case 'f':
		        fidNamesDownload=1;
			break;
		case 'r':
		        removeFiles=1;
			break;
		case 'R':
		        removeFiles=2;    /* don't ask confirmation questions */
			break;
		case 'a':
			host=strdup(optarg);
			break;
		case 'c':
			usercodeset=strdup(optarg);
			break;
		case 'e':
		        editPathStr1=strdup(optarg); 
			if (optind>argc-2) 
			  {usage(); return -1;}
			editPathStr2=strdup(argv[optind]); optind++; 
			editPathStr3=strdup(argv[optind]); optind++;
			break;
		case 'q':
		        squeezeRepeats=1;
			break;
		case 'p':
			psw=strdup(optarg);
			break;
		case 's':
			searchString=strdup(optarg);
			break;
		case 'u':
			tunePropsToUse=strdup(optarg);
			break;
		case 'x':
			taxiPropsToUse=strdup(optarg);
			break;
		case 'l':
			plstPropsToUse=strdup(optarg);
			break;
		default:
			fprintf(stderr, "unknown option %c\n", c);
			break;
		}
	}
	if(taxiPropsToUse==NULL) taxiPropsToUse=strdup(tunePropsToUse);
	if(plstPropsToUse==NULL) plstPropsToUse=strdup(taxiPropsToUse);

	lkerrAutoPrint(0);
	if(!host) {
		for(i=0; (i<=ssdpRetry)&&(host==NULL); i++) {
			lkerror();
			ssdp_discover(&host, &port);
			host?fprintf(stderr, "SSDP: %s:%i\n", host, port)
			    :fprintf(stderr, "SSDP: no answer\n");
		}
	}
	if (host==NULL) {
		lkerrorp("*", " -- Can't connect to Rio Karma");
		usage();
		if(psw) free(psw);
		if(searchString) free(searchString);
		if(usercodeset) free(usercodeset);
		return -1;
	}
	if (host!=NULL) {
	  karma=connectkarma(host);
	  lkerrorp("","");
	  free(host);
	}

	if(karma < 0) {
		usage();
		if(psw) free(psw);
		if(searchString) free(searchString);
		if(usercodeset) free(usercodeset);
		return -1;
	}
	
	currentKarma(1,&karma);
	
	/*init_properties("ISO8859-15");*/
	if(usercodeset != NULL){
		init_properties(usercodeset);
		if (lkerrorp("", "; Trying locale...")) {
			free(usercodeset);
			usercodeset = NULL;
		}
	}
	if(usercodeset == NULL){
		setlocale(LC_CTYPE, "");
		if(strlen(usercodeset = nl_langinfo(CODESET)) == 0) {
			fprintf(stderr, "* Warning: nl_langinfo problem; "
				"using default codeset %s\n", defaultCodeset);
			usercodeset = defaultCodeset;
		}
		init_properties(usercodeset);
 		if(! lkerrorp("",""))
		  fprintf(stderr, "* Using %s codeset.\n", usercodeset);
		/*free(usercodeset);*/
	}
	fprintf(stderr, "Using %s codeset.\n", usercodeset);
	lkerrAutoPrint(1);                    /* Setting auto-print-errors ON */
	
	if(psw){
		fprintf(stderr, "authenticate: %i\n", authenticate(karma, psw));
		free(psw);
	}else   fprintf(stderr, "authenticate: %i\n", authenticate(karma, ""));
	r = request_io_lock(karma,IO_LOCK_R);
	fprintf(stderr, "request_io_lock: %s\n", CHECK(r));

        get_device_settings(karma, &settings);
	dbChanged = cache_obsolete(settings);
        if(dbChanged) {
	  printf("Downloading Rio database... ");
	  fflush(stdout);
	  if(getUsedSize(karma)>3E9) printf(" please wait... ");
	  fflush(stdout);
	}
	free(settings);
	if(!r) load_database(karma);
       	if(dbChanged) printf("Done!\n");

        if(fidNamesDownload) {
	  fidNamesFileHandler(1, 0, NULL);
	  fidNamesFileHandler(3, 0, NULL);
	}

	if (searchString != NULL){
		vp=searchStr(searchString);
		free(searchString);
	}
	if(vp) downloadFids(vp, download, 
		     tunePropsToUse, taxiPropsToUse, plstPropsToUse, 
		     editPathStr1, editPathStr2, editPathStr3, squeezeRepeats);
	
	if(taxiPropsToUse) free(taxiPropsToUse); /*Enrique?*/
	if(plstPropsToUse) free(plstPropsToUse); /*Enrique?*/
	if(removeFiles>0 && vp) {
	  yesNot='Y';  for (i=0; vp[i]!=0; i++) continue;
	  if(i>0) {
	    printf("Ready to remove %d files; ", i);
	    if(removeFiles==1) {
	      printf("proceed (Y/N)? ");  scanf("%c", &yesNot);
	    }else{
	      printf("\n");
	    }
	    if(yesNot!='Y' && yesNot!='y') 
	      printf("Remove aborted!\n");
	    else {
	      fprintf(stderr, "request_io_lock (W): %s\n", 
		               CHECK(request_io_lock(karma,IO_LOCK_W)));
	      deleteFids(vp);
	    }
	  }
	  else fprintf(stderr, "No files to remove!\n");
	}

	if(optind < argc){
		fprintf(stderr, "request_io_lock: %s\n", 
			        CHECK(request_io_lock(karma,IO_LOCK_W)));
		if(synchronize_necessary(karma)){
			get_all_file_details(karma, &psw);
			import_properties(psw);
			free(psw);
		}
		fidNamesFileHandler(2, 0, NULL);
		for(; optind<argc; optind++)
			ftw(argv[optind], treecopy, 1);
	}
	fprintf(stderr, "release_io_lock: %s\n", CHECK(release_io_lock(karma)));
	fprintf(stderr, "hangup: %s\n", CHECK(hangup(karma)));

	free(vp);
	close(karma);
	save_properties();
	destroy_properties();
	fidNamesFileHandler(0, 0, NULL);
	return 0;
}
